unit Holiday;

interface

uses NDate;

function HolidayName(const Date: TDate): string;

implementation

uses SolarTerm;

type
  THoliday = record
    From, To_, Month, Day, OrdWday, Wday: Integer;
    SolarTerm: string;
    Name: string;
  end;

const
  FurikaeStart = 720360; // Date(1973, 4, 12).Ord
  KokuminStart = 724992; // Date(1985, 12, 17).Ord

  Holidays: array[0..37] of THoliday = (
    (From:1949;           Month: 1; Day: 1;            Name:''),
    (From:1949; To_:1999; Month: 1; Day:15;            Name:'l̓'),
    (From:2000;           Month: 1; OrdWday:2; Wday:0; Name:'l̓'),
    (From:1967;           Month: 2; Day:11;            Name:'LO̓'),
    (From:1949;           Month: 3; SolarTerm:'t';  Name:'t̓'),
    (From:1949; To_:1988; Month: 4; Day:29;            Name:'Vca'),
    (From:1989; To_:2006; Month: 4; Day:29;            Name:'݂ǂ̓'),
    (From:2007;           Month: 4; Day:29;            Name:'a̓'),
    (From:1949;           Month: 5; Day: 3;            Name:'@LO'),
    (From:2007;           Month: 5; Day: 4;            Name:'݂ǂ̓'),
    (From:1949;           Month: 5; Day: 5;            Name:'ǂ̓'),
    (From:1996; To_:2002; Month: 7; Day:20;            Name:'C̓'),
    (From:2003; To_:2019; Month: 7; OrdWday:3; Wday:0; Name:'C̓'),
    (From:2020; To_:2020; Month: 7; Day:23;            Name:'C̓'),
    (From:2021; To_:2021; Month: 7; Day:22;            Name:'C̓'),
    (From:2022;           Month: 7; OrdWday:3; Wday:0; Name:'C̓'),
    (From:2016; To_:2019; Month: 8; Day:11;            Name:'R̓'),
    (From:2020; To_:2020; Month: 8; Day:10;            Name:'R̓'),
    (From:2021; To_:2021; Month: 8; Day: 8;            Name:'R̓'),
    (From:2022;           Month: 8; Day:11;            Name:'R̓'),
    (From:1966; To_:2002; Month: 9; Day:15;            Name:'hV̓'),
    (From:2003;           Month: 9; OrdWday:3; Wday:0; Name:'hV̓'),
    (From:1949;           Month: 9; SolarTerm:'H';  Name:'H̓'),
    (From:1966; To_:1999; Month:10; Day:10;            Name:'̈̓'),
    (From:2000; To_:2019; Month:10; OrdWday:2; Wday:0; Name:'̈̓'),
    (From:2020; To_:2020; Month: 7; Day:24;            Name:'X|[c̓'),
    (From:2021; To_:2021; Month: 7; Day:23;            Name:'X|[c̓'),
    (From:2022;           Month:10; OrdWday:2; Wday:0; Name:'X|[c̓'),
    (From:1948;           Month:11; Day: 3;            Name:'̓'),
    (From:1948;           Month:11; Day:23;            Name:'ΘJӂ̓'),
    (From:1989; To_:2018; Month:12; Day:23;            Name:'Vca'),
    (From:2020;           Month: 2; Day:23;            Name:'Vca'),

    (From:1959; To_:1959; Month: 4; Day:10; Name:'cqmě̋V'),
    (From:1989; To_:1989; Month: 2; Day:24; Name:'aVc̑r̗'),
    (From:1990; To_:1990; Month:11; Day:12; Name:'ʗ琳a̋V'),
    (From:1993; To_:1993; Month: 6; Day: 9; Name:'cqmě̋V'),
    (From:2019; To_:2019; Month: 5; Day: 1; Name:'Vc̑ʂ̓'),
    (From:2019; To_:2019; Month:10; Day:22; Name:'ʗ琳a̋V')
  );

function _HolidayName(const Date: TDate): string;
var
  hd: THoliday;
  i, wd: Integer;
  r: Boolean;
begin
  Result := '';
  r := False;
  wd := Date.WeekDay;
  for i := 0 to High(Holidays) do begin
    hd := Holidays[i];
    if Date.Year < hd.From then
      Continue;
    if (hd.To_ > 0) and (Date.Year > hd.To_) then
      Continue;
    if (hd.Month > 0) and (Date.Month <> hd.Month) then
      Continue;
    if (hd.Day > 0) and (Date.Day <> hd.Day) then
      Continue;

    if hd.OrdWday > 0 then begin
      if hd.Wday <> wd then
        Continue;
      case hd.OrdWDay of
        1: r := Date.Day <= 7;
        2: r := (Date.Day >= 8) and (Date.Day <= 14);
        3: r := (Date.Day >= 15) and (Date.Day <= 21);
        4: r := (Date.Day >= 22) and (Date.Day <= 28);
        5: r := Date.Day >= 29;
      end;
    end else if hd.SolarTerm <> '' then
      r := SolarTermName(Date) = hd.SolarTerm
    else
      r := True;

    if r then begin
      Result := hd.Name;
      Break;
    end;
  end;
end;

function _IsHoliday(const Date: TDate): Boolean;
begin
  Result := _HolidayName(Date) <> '';
end;

function HolidayName(const Date: TDate): string;
var
  d, pd: TDate;
begin
  Result := _HolidayName(Date);

  if (Result = '') and (Date.Ord >= FurikaeStart) then begin
    d := Date;
    while True do begin
      pd := NDate.Date(d.Ord-1);
      if _IsHoliday(pd) then
      begin
        if pd.WeekDay = 6 then
        begin
          Result := 'U֋x';
          Break;
        end
      end else
        Break;
      d := pd;
    end;
  end;

  if (Result = '') and (Date.Ord >= KokuminStart) then
    if (Date.WeekDay <> 6) and
        _IsHoliday(NDate.Date(Date.Ord-1)) and
        _IsHoliday(NDate.Date(Date.Ord+1)) then
      Result := '̋x';
end;

end.
