unit SolarTerm;

interface

uses NDate;

function SolarTermName(const Date: TDate): string;

implementation

uses NMath;

const
  SolarTermNames: array[0..23] of string = (
    't', '', 'J', '', '', '䊎', 'Ď', '',
    '友', 'H', '', 'I', 'H', 'I', '~', '~',
    '', '', '~', '', '劦', 't', 'J', '['
  );

function JuliusYear(Year, Month, Day, Hour, Min, Sec: Integer;
    UTCOffset: Double): Double;
begin
  Dec(Year, 2000);
  if Month <= 2 then begin
    Dec(Year);
    Inc(Month, 12);
  end;
  Result := (365*Year + 30*Month + Day - 33.5 - UTCOffset/24 + Floor(3*(Month+1)/5) +
      Floor(Year / 4) - Floor(Year / 100) + Floor(Year / 400) +
      Hour / 24 + Min / 1440 + Sec / 86400 +
      (65 + Year) / 86400
      ) / 365.25;
end;

function SolarCelestialLongitude(JY: Double): Double;
begin
  Result := FMod(
      280.4603 + 360.00769 * JY +
      (1.9146 - 0.00005 * JY) * Sin((357.538 + 359.991 * JY) * Pi/180) +
      0.0200 * Sin((355.05 + 719.981 * JY) * Pi/180) +
      0.0048 * Sin((234.95 + 19.341 * JY) * Pi/180) +
      0.0020 * Sin((247.1 + 329.640 * JY) * Pi/180) +
      0.0018 * Sin((297.8 + 4452.67  * JY) * Pi/180) +
      0.0018 * Sin((251.3 + 0.20  * JY) * Pi/180) +
      0.0015 * Sin((343.2 + 450.37 * JY) * Pi/180) +
      0.0013 * Sin((81.4 + 225.18 * JY) * Pi/180) +
      0.0008 * Sin((132.5 + 659.29 * JY) * Pi/180) +
      0.0007 * Sin((153.3 + 90.38 * JY) * Pi/180) +
      0.0007 * Sin((206.8 + 30.35 * JY) * Pi/180) +
      0.0006 * Sin((29.8 + 337.18 * JY) * Pi/180) +
      0.0005 * Sin((207.4 + 1.50 * JY) * Pi/180) +
      0.0005 * Sin((291.2 + 22.81 * JY) * Pi/180) +
      0.0004 * Sin((234.9 + 315.56 * JY) * Pi/180) +
      0.0004 * Sin((157.3 + 299.30 * JY) * Pi/180) +
      0.0004 * Sin((21.1 + 720.02 * JY) * Pi/180) +
      0.0003 * Sin((352.5 + 1079.97 * JY) * Pi/180) +
      0.0003 * Sin((329.7 + 44.43 * JY) * Pi/180),
      360
      );
end;

function SolarTermName(const Date: TDate): string;
var
  jy: Double;
  t1, t2: Integer;
begin
  if (Date.Year = 2030) and (Date.Month = 2) then begin
    if Date.Day = 18 then begin
      Result := '';
      Exit;
    end;
    if Date.Day = 19 then begin
      Result := SolarTermNames[22];
      Exit;
    end;
  end;

  jy := JuliusYear(Date.Year, Date.Month, Date.Day, 0, 0, 0, 9);
  t1 := Trunc(SolarCelestialLongitude(jy) / 15);
  t2 := Trunc(SolarCelestialLongitude(jy + 1/365.25) / 15);
  if t1 = t2 then
    Result := ''
  else
    Result := SolarTermNames[t2];
end;

end.
